def print_msg_box(msg, indent=1, width=None, title=None):
    """Print message-box with optional title."""
    lines = msg.split('\n')
    space = " " * indent
    if not width:
        width = max(map(len, lines))
    box = f'╔{"═" * (width + indent * 2)}╗\n'  # upper_border
    if title:
        box += f'║{space}{title:<{width}}{space}║\n'  # title
        box += f'║{space}{"-" * len(title):<{width}}{space}║\n'  # underscore
    box += ''.join([f'║{space}{line:<{width}}{space}║\n' for line in lines])
    box += f'╚{"═" * (width + indent * 2)}╝'  # lower_border
    print(box)



from tabulate import tabulate
from .__init__ import print_msg_box


class error(Exception):
    pass

class Stack(object):
    """
    Stack is an abstract data type used to store
    data in Last In First Out method (LIFO).

    We are going to use list(array) to build the stack

    Main feature of this array based stack is

    * Insert
    * deletion of last element
    * Finding the length of the stack
    * Check is the stack is empty
    * get the element at the top of the stack

    Example for application using stack:
       1. browser fetching the previously used URL using back
          and forward button
       2. stack of plates in a restaurant
       3. Validating html file by matching the brackets
       4. etc ...

    Reference:
         https://en.wikipedia.org/wiki/Stack_(abstract_data_type)

    Usage:
    >>> from stack import Stack

    >>> my_stack = Stack(5)

    >>> for value in [10,1,24,200,1000]:
    ...    my_stack.push(value)

    >>> my_stack.pretty_print()
    |-----:|
    | 1000 |
    |  200 |
    |   24 |
    |    1 |
    |   10 |

    >>> my_stack.pop()
    1000

    >>> my_stack.pretty_print()
    |----:|
    | 200 |
    |  24 |
    |   1 |
    |  10 |
    >>> my_stack.top()
    200

    >>> len(my_stack)
    4

    >>> my_stack.push(10)
    >>> my_stack.push(1)
    Traceback (most recent call last):
       ...
    stack.error: Stack size reached maximum

    >>> len(my_stack)
    5
    """


    def __init__(self, size = 0):
        """ Initilise the stack with size 0 and empty data """
        self.data = []
        self.size = size

    def __len__(self):
        """Returns the length of the data"""
        return len(self.data)

    def is_empty(self):
        """Returns boolean False if the data is not empty else True"""
        return len(self.data) == 0

    def push(self, value):
        """Append the value to data """
        if len(self.data) == self.size:
            """print the stack"""
            raise error("Stack size reached maximum")

        self.data.append(value)


    def pop(self):
        """ Return value from data as well as delete the value
        from data after return"""
        if self.is_empty():
            raise error("Stack is empty")
        else:
            return self.data.pop()

    def top(self):
        """Return the last pushed value from data """
        if self.is_empty():
            raise error("Stack is empty")
        else:
            return self.data[-1]

    def print_hint(self):
        """Print the documentation for class stack"""
        print(self.__doc__)

    def pretty_print(self):
        """ Print the stack """
        list_of_list = []
        for value in self.data:
            list_of_list.append([value])
        table_data = list_of_list[::-1]
        print(tabulate(table_data, tablefmt="pipe"))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from .__init__ import print_msg_box

class Queue(object):
    """
    Queue is an abstract data type used to store
    data in  first-in-first-out (FIFO) data structure.

    We are going to use list(array) to build the queue

    Main feature of this array based queue is

    * enqueue(Addition of an element at the last index of array)
    * dequeue(Popping the element from the first index of the array)
    * Finding the length of the queue
    * Check is the queue is empty
    * get the element at the front of the queue

    Example for application using queue:
       1. Priority queue usages
       2. Any line on reservation counters
       3. etc ...

    Reference:
         https://en.wikipedia.org/wiki/Queue_(abstract_data_type)

    Usage:
    >>> from queue import Queue

    >>> sample_queue = Queue(5)

    >>> for value in [7,17,367,27,777]:
    ...     sample_queue.enqueue(value)

    >>> sample_queue.print_queue()
    | 7 |
    | 17 |
    | 367 |
    | 27 |
    | 777 |

    >>> sample_queue.dequeue()
    7

    >>> sample_queue.print_queue()
    | 17 |
    | 367 |
    | 27 |
    | 777 |

    >>> sample_queue.front()
    17

    >>> len(sample_queue)
    4

    >>> sample_queue.enqueue(77)

    >>> sample_queue.print_queue()
    | 17 |
    | 367 |
    | 27 |
    | 777 |
    | 77 |
    >>> sample_queue.enqueue(77777)
    Traceback (most recent call last):
    ...
    Exception: Queue size limit reached maximum

    >>> len(sample_queue)
    5
    """
    def __init__(self, size = 0):
        """ Initilise the queue with size 0 and empty data """
        self.data = []
        self.size = size

    def __len__(self):
        """Returns the length of the queue"""
        return len(self.data)

    def is_empty(self):
        """Returns boolean True if queue is empty and False if we have some data present"""
        return len(self.data) == 0

    def enqueue(self, value):
        """Enqueue (append) the value to queue """
        if len(self.data) == self.size:
            """Current queue list:"""
            self.print_queue()
            raise Exception("Queue size limit reached maximum")

        self.data.append(value)

    def dequeue(self):
        """ The front element is popped from the Queue and value of the element is returned"""
        if self.is_empty():
            raise Exception("Queue is empty !!! Please add data to the Queue :) ")
        else:
            return self.data.pop(0)

    def front(self):
        """Return the first element present in the queue """
        if self.is_empty():
            raise Exception("Queue is empty !!! Please add data to the Queue :) ")
        else:
            return self.data[0]

    def print_hint(self):
        """Print the details related to Queue and its implementation"""
        print(self.__doc__)

    def print_queue(self):
        """ Print the current queue"""
        for value in self.data:
            element = f'| {value} |'
            print(element)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from .__init__ import print_msg_box

class BinarySearyAlgorithm:

    #Define the bsearch method i.e. binary search
    def bsearch(self, list, val):

        list_size = len(list) - 1

        low = 0 #lower index
        high = list_size

    # Find the middle most value

        while low <= high:
            mid = (low + high)// 2 #divide

            if list[mid] == val:
                return mid
    # Compare the value the middle most value
            if val > list[mid]:
                low = mid + 1
            else:
                high = mid - 1
    # Value isn't found anywhere in the list, then return none            
        if low > high:
            return None

    def print_bsearch_hint(self):
        message = """
        FIND AN ELEMENT FROM A GIVEN SORTED LIST USING BINARY SEARCH ALGORITHM AND PRINT THE ELEMENT'S POSITION
        -+-+-+-
        In binary search we take a sorted list of elements and start looking for an element at the middle of the list.
        If the search value matches with the middle value in the list we complete the search.
        Otherwise we eleminate half of the list of elements by choosing whether to procees with the right or left half
        of the list depending on the value of the item searched. This is possible as the list is sorted and it is
        much quicker than linear search. Here we divide the given list and conquer by choosing the proper half of the
        list. We repeat this approcah till we find the element or conclude about it's absence in the list.
        -+-+-+-
        PSEUDOCODE:
            1. Compare 'x' with the middle element
            2. If x matches with the middle most element, we return the index of the middle element i.e. 'mid'.
            3. Else if x is greater than the middle element, x will lie in the right sub-part of the list from the middle element.Thus, we recur the right part of the list.
            4. Else, the x is smaller than the middle element, so we recur the left sub-part of the list.
        -+-+-+-
        TIME COMPLEXITY OF BINARY SEARCH ALGORITHM:

            Worst-case performance: O(log n)
            Best-case performance: O(1)
            Average performance: O(log n)
            Worst-case space complexity: O(1)
        -+-+-+-
        MORE INFO HERE: https://en.wikipedia.org/wiki/Binary_search_algorithm
        """
        print_msg_box(message)

# Initialize the sorted list

#list = [2,7,19,34,53,72] #samplelist

# Print the search result
#obj1 = BinarySearyAlgorithm()
#print(obj1.bsearch(list,72))
#print(bsearch(list,100))
#print(obj1.print_bsearch_hint())



from .__init__ import print_msg_box
import collections
class Node:
    def __init__(self, val):

        self.left = None
        self.right = None
        self.val = val

    def insert(self, val):
        if self.val:
            if val < self.val:
                if self.left is None:
                    self.left = Node(val)
                else:
                    self.left.insert(val)
            else:
                if self.right is None:
                    self.right = Node(val)
                else:
                    self.right.insert(val)
        else:
            self.val = val

class BinaryTreeAlgorithms:
    def print_tree(self,root,hint=False):
        if(hint==True):
            self.print_tree_hint()
        if root == None:
            return
        print(root.val,end =":")
        if root.left:
            print("L: {} ,".format(root.left.val),end ="")
        if root.right:
            print("R: {} ,".format(root.right.val),end="")
        print("\n")
        self.print_tree(root.left)
        self.print_tree(root.right)

    def print_tree_hint(self):
        message = """
        Printing A Binary Tree
        ------------------------------------

        Purpose : Printing a Binary Tree(Both Left and Right Node)
        Method : Recursion, Binary Tree

        Time Complexity : Worst Case - O(n), n = Number of nodes in a Binary Tree

        Hint :
        Print the root, use recursion and call into the left and right

        Pseudocode :
        --> if(root == None) return
        --> print(root.value)
        --> print(root.left.value)
        --> print(root.right.value)
        --> make recursion calls,
            print_tree(root.left)
            print.tree(root.right)

        Visualization:

        Given Binary Tree :

                            +------+
                            |  12  |         <-- root
                            +------+
                            /      \\
                           /        \\
                   +------+          +------+
    root.left -->  |  6   |          |  14  |   <-- root.right
                   +------+          +------+
                   /      \\          /      \\
                  /        \\        /        \\
          +------+     +------+   +------+    +------+
          |  3   |     |  7   |   |  12  |    |  15  |
          +------+     +------+   +------+    +------+


           Step 1: print root, root.left and root.right

                12 : L:6, R: 14

           Step 2 : call recursive functions

           f(root.left) :

                           +------+
                           |  6   |    <-- root
                           +------+
                           /      \\
                          /        \\
                  +------+         +------+
   root.left -->  |  3   |         |  7   |  <-- root.right
                  +------+         +------+

          +------------------------+
          | Repeat Step 1 & Step 2 |   <-- recursion calls
          +------------------------+

          f(root.right) :

                          +------+
                          |  14  |    <-- root
                          +------+
                          /      \\
                         /        \\
                 +------+         +------+
  root.left -->  |  12  |         |  15  |  <-- root.right
                 +------+         +------+

         +------------------------+
         | Repeat Step 1 & Step 2 |      <-- recursion calls
         +------------------------+

         Finally The Output :

         -------------------------
        |    12:L: 6 ,R: 14 ,     |
        |                         |
        |    6:L: 3 ,R: 7 ,       |
        |                         |
        |    3:                   |
        |                         |
        |    7:                   |
        |                         |
        |    14:L: 12 ,R: 15 ,    |
        |                         |
        |    12:                  |
        |                         |
        |    15:                  |
         -------------------------

        Learn More:
        - Binary Trees - https://en.wikipedia.org/wiki/Binary_tree
        - Recursion - https://en.wikipedia.org/wiki/Recursion_(computer_science)
        """
        print_msg_box(message)
    
    def Inorder_print(self,root,hint=False):
        if(hint==True):
            self.Inorder_print_hint()
        if root == None:
            return
        self.Inorder_print(root.left)
        print(root.val,end =", ")
        self.Inorder_print(root.right)

    def Inorder_print_hint(self):
        message = """
        Printing A Binary Tree InOrder Traversal
        ------------------------------------

        Purpose : Printing a Binary Tree(InOrder Traversal)
        Method : Recursion, Binary Tree

        Time Complexity : Worst Case - O(n), n = Number of nodes in a Binary Tree

        Hint :
        print order ->  LEFT -- ROOT -- RIGHT
        Call into left recursively till exist,Print the root, use recursion to call into the right

        Pseudocode :
        --> if(root == None) return
        --> print(root.left.value)
        --> print(root.value)
        --> print(root.right.value)

        Visualization:

        Given Binary Tree :

                            +------+
                            |  12  |         <-- root
                            +------+
                            /      \\
                           /        \\
                   +------+          +------+
    root.left -->  |  6   |          |  14  |   <-- root.right
                   +------+          +------+
                   /      \\          /      \\
                  /        \\        /        \\
          +------+     +------+   +------+    +------+
          |  3   |     |  7   |   |  13  |    |  15  |
          +------+     +------+   +------+    +------+


        step : call recursive functions on root.left, print root, call on right

                f(root.left) :

                           +------+
                           |  3   |    <-- root
                           +------+
                           /      \\
                          /        \\
                  +------+         +------+
   root.left -->  | None |         | None |  <-- root.right
                  +------+         +------+

                output : LEFT -- ROOT -- RIGHT 
                         None     3      None
            

                           +------+
                           |  6   |    <-- root
                           +------+
                           /      \\
                          /        \\
                  +------+         +------+
   root.left -->  |  3   |         |  7   |  <-- root.right
                  +------+         +------+

                output : LEFT -- ROOT -- RIGHT 
                          3       6        7

          f(root.right) :

                          +------+
                          |  14  |    <-- root
                          +------+
                          /      \\
                         /        \\
                 +------+         +------+
  root.left -->  |  13  |         |  15  |  <-- root.right
                 +------+         +------+

               output : LEFT -- ROOT -- RIGHT 
                         13      14      15

         Finally The Output :

         ---------------------------------
        |    3, 6, 7, 12, 13, 14, 15,     |
         ---------------------------------

        Learn More:
        - Binary Trees - https://en.wikipedia.org/wiki/Binary_tree
        - Recursion - https://en.wikipedia.org/wiki/Recursion_(computer_science)
        """
        print_msg_box(message)

    def Preorder_print(self,root,hint=False):
        if(hint==True):
            self.Preorder_print_hint()
        if root == None:
            return
        print(root.val,end =", ")
        self.Preorder_print(root.left)
        self.Preorder_print(root.right)

    def Preorder_print_hint(self):
        message = """
        Printing A Binary Tree PreOrder Traversal
        ------------------------------------

        Purpose : Printing a Binary Tree(PreOrder Traversal)
        Method : Recursion, Binary Tree

        Time Complexity : Worst Case - O(n), n = Number of nodes in a Binary Tree

        Hint :
        print order ->  ROOT -- LEFT -- RIGHT
        Print the root, use recursion to call into the left and the right subtree

        Pseudocode :
        --> if(root == None) return
        --> print(root.value)
        --> print(root.left.value)
        --> print(root.right.value)

        Visualization:

        Given Binary Tree :

                                +------+
                                |  12  |   <-- root
                                +------+
                                /      \\
                               /        \\
                        +------+          +------+
         root.left -->  |  6   |          |  14  |   <-- root.right
                        +------+          +------+
                        /      \\          /      \\
                       /        \\        /        \\
                +------+     +------+   +------+    +------+
                |  3   |     |  7   |   |  13  |    |  15  |
                +------+     +------+   +------+    +------+


        step 1 : Print the root value
                    
                                +------+
                                |  6   |    <-- root
                                +------+
                                /      \\
                               /        \\
                        +------+         +------+
         root.left -->  |  3   |         |  7   |  <-- root.right
                        +------+         +------+

                output : 6

                f(root.left) :

                                +------+
                                |  3   |    <-- root
                                +------+
                                /      \\
                               /        \\
                        +------+         +------+
         root.left -->  | None |         | None |  <-- root.right
                        +------+         +------+

                output : ROOT -- LEFT -- RIGHT 
                          3      None     None

          f(root.right) :

                                +------+
                                |  14  |    <-- root
                                +------+
                                /      \\
                               /        \\
                        +------+         +------+
         root.left -->  |  13  |         |  15  |  <-- root.right
                        +------+         +------+

               output : ROOT -- LEFT -- RIGHT 
                         14      13      15

         Finally The Output :

         ---------------------------------
        |    12, 6, 3, 7, 14, 13, 15,     |
         ---------------------------------

        Learn More:
        - Binary Trees - https://en.wikipedia.org/wiki/Binary_tree
        - Recursion - https://en.wikipedia.org/wiki/Recursion_(computer_science)
        """
        print_msg_box(message)

    def Postorder_print(self,root,hint=False):
        if(hint==True):
            self.Postorder_print_hint()
        if root == None:
            return
        self.Postorder_print(root.left)
        self.Postorder_print(root.right)
        print(root.val,end =", ")

    def Postorder_print_hint(self):
        message = """
        Printing A Binary Tree PostOrder Traversal
        ------------------------------------

        Purpose : Printing a Binary Tree(PostOrder Traversal)
        Method : Recursion, Binary Tree

        Time Complexity : Worst Case - O(n), n = Number of nodes in a Binary Tree

        Hint :
        print order ->  LEFT -- RIGHT -- ROOT
        use recursion to call into the left and the right subtree, Print the root

        Pseudocode :
        --> if(root == None) return
        --> print(root.left.value)
        --> print(root.right.value)
        --> print(root.value)

        Visualization:

        Given Binary Tree :

                                +------+
                                |  12  |   <-- root
                                +------+
                                /      \\
                               /        \\
                        +------+          +------+
         root.left -->  |  6   |          |  14  |   <-- root.right
                        +------+          +------+
                        /      \\          /      \\
                       /        \\        /        \\
                +------+     +------+   +------+    +------+
                |  3   |     |  7   |   |  13  |    |  15  |
                +------+     +------+   +------+    +------+


        step 1 : Print the left, print right, print root
                    
                                +------+
                                |  6   |    <-- root
                                +------+
                                /      \\
                               /        \\
                        +------+         +------+
         root.left -->  |  3   |         |  7   |  <-- root.right
                        +------+         +------+


                output : LEFT -- RIGHT -- ROOT 
                          3        7        6


         Finally The Output :

         ---------------------------------
        |    3, 7, 6, 13, 15, 14, 12     |
         ---------------------------------

        Learn More:
        - Binary Trees - https://en.wikipedia.org/wiki/Binary_tree
        - Recursion - https://en.wikipedia.org/wiki/Recursion_(computer_science)
        """
        print_msg_box(message)

    def rangeSumBST(self, root,L,R):
        if(root == None):
            return 0
        sum1 = 0; sum2 = 0
        if(root.left):
            sum1 = self.rangeSumBST(root.left,L,R)
        if(root.right):
            sum2 = self.rangeSumBST(root.right,L,R)
        if((root.val >= L )and (root.val <= R)):
            return root.val + sum1 + sum2
        else:
            return sum1 + sum2

    def mergeTrees(self, t1, t2):
        if(t1 == None and t2 == None):
            return None
        if(t2 == None):
            return t1
        if(t1 == None):
            return t2
        t1.val = t1.val + t2.val
        t1.left = self.mergeTrees(t1.left,t2.left)
        t1.right = self.mergeTrees(t1.right,t2.right)
        return t1

    def sumOfLeftLeaves(self, root):
        if(root == None):
            return 0
        sum = 0
        if(root.left != None and (root.left.left == None and root.left.right == None)):
            sum = root.left.val
        return sum + self.sumOfLeftLeaves(root.left)+self.sumOfLeftLeaves(root.right)

    def isSameTree(self, p, q):
        if(p == None and q == None):
            return True
        if(p == None or q == None):
            return False
        return (p.val == q.val) and (self.isSameTree(p.left,q.left)) and (self.isSameTree(p.right,q.right))

    def isCousins(self, root, x, y):
        if not root:
            return root
        queue = collections.deque([root])
        level = set()
        last = n_last = root
        while any(queue):
            node = queue.popleft()
            same_father = set()
            if node.left:
                level.add(node.left.val)
                queue.append(node.left)
                n_last = node.left
                same_father.add(node.left.val)
            if node.right:
                level.add(node.right.val)
                queue.append(node.right)
                n_last = node.right
                same_father.add(node.right.val)
            if x in same_father and y in same_father:
                return False
            if node == last:
                last = n_last
                if x in level and y in level:
                    return True
                level = set()
        return False

    def countNode(self,root):
        '''
        countNode will take root node 
        as input and return the number 
        node present in the tree.
        '''
        if (root == None):
            return 0
        return 1 + self.countNode(root.left) + self.countNode(root.right)

    def KthLevel(self,root, k):
        '''
        KthLevel will take root node and level value (k),
        and print the element present at that level 
        Note: level for the root node is 1
        '''
        if (root == None):
            return
        if (k ==1):
            print(root.data,end=', ')
            return
        self.KthLevel(root.left, k-1)
        self.KthLevel(root.right, k-1)
        return


#root1 = Node(12)
#root1.insert(6)
#root1.insert(14)
#root1.insert(3)
#root1.insert(7)
#root1.insert(15)
#root1.insert(13)
# ping = BinaryTreeAlgorithms()
# ping.Inorder_print(root1)
# ping.Preorder_print(root1)
# ping.Postorder_print(root1)


